""" Monkey patches for loopy.symbolic

Use this module to insert pymbolic nodes and the likes.
"""
from dune.perftool.error import PerftoolError
from pymbolic.mapper.substitutor import make_subst_func

import loopy as lp
import pymbolic.primitives as prim
from pytools import ImmutableRecord
from six.moves import intern


#
# Pymbolic nodes to insert into the symbolic language understood by loopy
#


class SumfactKernel(ImmutableRecord, prim.Variable):
    def __init__(self,
                 a_matrices=None,
                 buffer=None,
                 stage=1,
                 preferred_position=None,
                 restriction=0,
                 within_inames=(),
                 input=None,
                 padding=frozenset(),
                 index=None,
                 insn_dep=frozenset(),
                 coeff_func=None,
                 element=None,
                 component=None,
                 ):
        # Check the input and apply defaults where necessary
        assert isinstance(a_matrices, tuple)
        from dune.perftool.sumfact.amatrix import AMatrixBase
        assert all(isinstance(m, AMatrixBase) for m in a_matrices)

        assert stage in (1, 3)

        if preferred_position is not None:
            assert isinstance(preferred_position, int)

        if stage == 3:
            assert isinstance(restriction, tuple)

        assert isinstance(within_inames, tuple)

        assert isinstance(insn_dep, frozenset)

        ImmutableRecord.__init__(self,
                                 a_matrices=a_matrices,
                                 buffer=buffer,
                                 stage=stage,
                                 preferred_position=preferred_position,
                                 restriction=restriction,
                                 within_inames=within_inames,
                                 input=input,
                                 padding=padding,
                                 index=index,
                                 insn_dep=insn_dep,
                                 coeff_func=coeff_func,
                                 element=element,
                                 component=component,
                                 )

        prim.Variable.__init__(self, "SUMFACT")

    #
    # The methods/fields needed to get a well-formed pymbolic node
    #
    def __getinitargs__(self):
        return (self.a_matrices, self.buffer, self.stage, self.preferred_position, self.restriction, self.within_inames, self.input, self.padding, self.index, self.insn_dep, self.coeff_func, self.element, self.component)

    def stringifier(self):
        return lp.symbolic.StringifyMapper

    init_arg_names = ("a_matrices", "buffer", "stage", "preferred_position", "restriction", "within_inames", "input", "padding", "index", "insn_dep", "coeff_func", "element", "component")

    mapper_method = "map_sumfact_kernel"

    #
    # Some convenience methods to extract information about the sum factorization kernel
    #

    @property
    def length(self):
        return len(self.a_matrices)

    @property
    def vectorized(self):
        return next(iter(self.a_matrices)).vectorized

    @property
    def cache_key(self):
        """ The cache key that can be used in generation magic,
        Any two sum factorization kernels having the same cache_key
        are realized simulatneously!
        """
        return hash((self.a_matrices, self.restriction, self.stage, self.buffer))

    @property
    def flat_input_shape(self):
        """ The 'flat' input tensor shape """
        from pytools import product
        shape = (product(mat.cols for mat in self.a_matrices),)
        if self.vectorized:
            shape = shape + (4,)
        return shape


class FusedMultiplyAdd(prim.Expression):
    """ Represents an FMA operation """

    init_arg_names = ("mul_op1", "mul_op2", "add_op")

    def __init__(self, mul_op1, mul_op2, add_op):
        self.mul_op1 = mul_op1
        self.mul_op2 = mul_op2
        self.add_op = add_op

    def __getinitargs__(self):
        return (self.mul_op1, self.mul_op2, self.add_op)

    def stringifier(self):
        return lp.symbolic.StringifyMapper

    mapper_method = intern("map_fused_multiply_add")


#
# Mapper methods to monkey patch into the visitor base classes!
#


def identity_map_sumfact_kernel(self, expr, *args):
    return expr


def walk_map_sumfact_kernel(self, expr, *args):
    self.visit(expr)


def stringify_map_sumfact_kernel(self, expr, *args):
    return "SUMFACT"


def dependency_map_sumfact_kernel(self, expr):
    return set()


def needs_resolution(self, expr):
    raise PerftoolError("SumfactKernel node is a placeholder and needs to be removed!")


def identity_map_fused_multiply_add(self, expr, *args):
    return FusedMultiplyAdd(self.rec(expr.mul_op1, *args),
                            self.rec(expr.mul_op2, *args),
                            self.rec(expr.add_op, *args),
                            )


def walk_map_fused_multiply_add(self, expr, *args):
    if not self.visit(expr):
        return

    self.rec(expr.mul_op1, *args)
    self.rec(expr.mul_op2, *args)
    self.rec(expr.add_op, *args)


def stringify_map_fused_multiply_add(self, expr, enclosing_prec):
    from pymbolic.mapper.stringifier import PREC_NONE
    return "fma(%s*%s+%s)" % (self.rec(expr.mul_op1, PREC_NONE),
                              self.rec(expr.mul_op2, PREC_NONE),
                              self.rec(expr.add_op, PREC_NONE))


def dependency_map_fused_multiply_add(self, expr):
    return self.combine((self.rec(expr.mul_op1),
                         self.rec(expr.mul_op2),
                         self.rec(expr.add_op)
                         ))


def type_inference_fused_multiply_add(self, expr):
    return self.rec(expr.mul_op1)


def vectorizability_map_fused_multiply_add(self, expr):
    return all((self.rec(expr.mul_op1), self.rec(expr.mul_op2), self.rec(expr.add_op)))


#
# Do the actual monkey patching!!!
#

# SumfactKernel node
lp.symbolic.IdentityMapper.map_sumfact_kernel = identity_map_sumfact_kernel
lp.symbolic.SubstitutionMapper.map_sumfact_kernel = lp.symbolic.SubstitutionMapper.map_variable
lp.symbolic.WalkMapper.map_sumfact_kernel = walk_map_sumfact_kernel
lp.symbolic.StringifyMapper.map_sumfact_kernel = stringify_map_sumfact_kernel
lp.symbolic.DependencyMapper.map_sumfact_kernel = dependency_map_sumfact_kernel
lp.target.c.codegen.expression.ExpressionToCExpressionMapper.map_sumfact_kernel = needs_resolution
lp.type_inference.TypeInferenceMapper.map_sumfact_kernel = needs_resolution

# FusedMultiplyAdd node
lp.symbolic.IdentityMapper.map_fused_multiply_add = identity_map_fused_multiply_add
lp.symbolic.WalkMapper.map_fused_multiply_add = walk_map_fused_multiply_add
lp.symbolic.StringifyMapper.map_fused_multiply_add = stringify_map_fused_multiply_add
lp.symbolic.DependencyMapper.map_fused_multiply_add = dependency_map_fused_multiply_add
lp.type_inference.TypeInferenceMapper.map_fused_multiply_add = type_inference_fused_multiply_add
lp.expression.VectorizabilityChecker.map_fused_multiply_add = vectorizability_map_fused_multiply_add

#
# Some helper functions!
#


def substitute(expr, replacemap):
    """ A replacement for pymbolic.mapper.subsitutor.substitute which is aware of all
    monkey patches etc.
    """
    return lp.symbolic.SubstitutionMapper(make_subst_func(replacemap))(expr)
