import copy
import itertools

from dune.perftool.loopy.symbolic import substitute
from dune.perftool.pdelab.argument import (name_accumulation_variable,
                                           name_coefficientcontainer,
                                           pymbolic_coefficient,
                                           PDELabAccumulationFunction,
                                           )
from dune.perftool.generation import (backend,
                                      barrier,
                                      built_instruction,
                                      domain,
                                      dump_accumulate_timer,
                                      function_mangler,
                                      generator_factory,
                                      get_counted_variable,
                                      get_counter,
                                      get_global_context_value,
                                      globalarg,
                                      iname,
                                      instruction,
                                      post_include,
                                      kernel_cached,
                                      retrieve_cache_items,
                                      silenced_warning,
                                      temporary_variable,
                                      transform,
                                      )
from dune.perftool.options import get_option
from dune.perftool.loopy.flatten import flatten_index
from dune.perftool.loopy.buffer import (get_buffer_temporary,
                                        initialize_buffer,
                                        switch_base_storage,
                                        )
from dune.perftool.sumfact.quadrature import nest_quadrature_loops
from dune.perftool.pdelab.localoperator import determine_accumulation_space
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.pdelab.signatures import assembler_routine_name
from dune.perftool.pdelab.spaces import (name_lfs,
                                         name_lfs_bound,
                                         )
from dune.perftool.pdelab.geometry import (local_dimension,
                                           world_dimension,
                                           )
from dune.perftool.sumfact.amatrix import (AMatrix,
                                           LargeAMatrix,
                                           quadrature_points_per_direction,
                                           basis_functions_per_direction,
                                           construct_amatrix_sequence,
                                           )
from dune.perftool.sumfact.switch import (get_facedir,
                                          get_facemod,
                                          )
from dune.perftool.loopy.symbolic import SumfactKernel
from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.tools import get_pymbolic_basename
from dune.perftool.error import PerftoolError
from pymbolic.primitives import (Call,
                                 Product,
                                 Subscript,
                                 Variable,
                                 )
from dune.perftool.sumfact.quadrature import quadrature_inames
from dune.perftool.sumfact.vectorization import find_sumfact
from loopy.symbolic import FunctionIdentifier, IdentityMapper
from dune.perftool.sumfact.permutation import (_sf_permutation_strategy,
                                               _permute_backward,
                                               _permute_forward,
                                               )

import loopy as lp
import numpy as np
import pymbolic.primitives as prim
from pytools import product


@function_mangler
def horizontal_add_function_mangler(knl, func, arg_dtypes):
    if func == "horizontal_add":
        from dune.perftool.loopy.vcl import get_vcl_type
        vcl = lp.types.NumpyType(get_vcl_type(np.float64))
        return lp.CallMangleInfo("horizontal_add", (lp.types.NumpyType(np.float64),), (vcl,))


@iname
def _sumfact_iname(bound, _type, count):
    name = "sf_{}_{}".format(_type, str(count))
    domain(name, bound)
    return name


def sumfact_iname(bound, _type):
    count = get_counter('_sumfac_iname_{}'.format(_type))
    name = _sumfact_iname(bound, _type, count)
    return name


@kernel_cached
def accum_iname(restriction, bound, i):
    return sumfact_iname(bound, "accum")


def setup_theta(inp, element, restriction, component, index, coeff_func):
    if index is None:
        index = ()
    else:
        index = (index,)
    # Write initial coefficients into buffer
    lfs = name_lfs(element, restriction, component)
    basisiname = sumfact_iname(name_lfs_bound(lfs), "basis")
    container = coeff_func(restriction)
    coeff = pymbolic_coefficient(container, lfs, basisiname)
    assignee = Subscript(Variable(inp), (Variable(basisiname),) + index)
    return instruction(assignee=assignee,
                       expression=coeff,
                       )


def name_test_function_contribution(test):
    count = get_counter('sumfact_contribution_counter')
    grad = "grad_" if test.reference_grad else ""
    return restricted_name("contrib_{}phi_{}".format(grad, str(count)), test.restriction)


@backend(interface="accum_insn", name="sumfact")
def generate_accumulation_instruction(visitor, accterm, measure, subdomain_id):
    pymbolic_expr = visitor(accterm.term)

    if pymbolic_expr == 0:
        return

    dim = world_dimension()

    # If this is a gradient, we find the gradient iname
    additional_inames = frozenset({})
    if accterm.new_indices is not None:
        for i in accterm.new_indices:
            if i not in visitor.dimension_indices:
                from dune.perftool.pdelab.localoperator import grad_iname
                additional_inames = additional_inames.union(frozenset({grad_iname(i, dim)}))

    def emit_sumfact_kernel(i, restriction, insn_dep):
        # Construct the matrix sequence for this sum factorization
        a_matrices = construct_amatrix_sequence(transpose=True,
                                                derivative=i if accterm.new_indices else None,
                                                facedir=get_facedir(accterm.argument.restriction),
                                                facemod=get_facemod(accterm.argument.restriction),
                                                )

        sf = SumfactKernel(a_matrices=a_matrices,
                           restriction=(accterm.argument.restriction, restriction),
                           stage=3,
                           preferred_position=i if accterm.new_indices else None,
                           within_inames=visitor.inames,
                           )

        # TODO: Move this away!
        from dune.perftool.sumfact.vectorization import attach_vectorization_info
        sf = attach_vectorization_info(sf)

        # Extract again, for compatibility
        # TODO away!
        a_matrices = sf.a_matrices
        buf = sf.buffer
        inp = sf.input
        index = sf.index
        padding = sf.padding

        if buf is None:
            buf = get_counted_variable("buffer")
        if inp is None:
            inp = get_counted_variable("input")

        # Initialize a base storage for this buffer and get a temporay pointing to it
        shape = tuple(mat.cols for mat in a_matrices if mat.face is None)
        dim_tags = ",".join(['f'] * local_dimension())
        if index is not None:
            shape = shape + (4,)
            dim_tags = dim_tags + ",c"
            index = (index,)
            vectag = frozenset({"gradvec"})
        else:
            index = ()
            vectag = frozenset()

        temp = initialize_buffer(buf).get_temporary(shape=shape,
                                                 dim_tags=dim_tags,
                                                 name=inp,
                                                 )

        # Those input fields, that are padded need to be set to zero
        # in order to do a horizontal_add later on
        for pad in padding:
            assignee = prim.Subscript(prim.Variable(temp),
                                      tuple(Variable(i) for i in quadrature_inames()) + (pad,))
            instruction(assignee=assignee,
                        expression=0,
                        forced_iname_deps=frozenset(quadrature_inames() + visitor.inames),
                        forced_iname_deps_is_final=True,
                        tags=frozenset(["quadvec", "gradvec"])
                        )

        # Replace gradient iname with correct index for assignement
        replace_dict = {}
        for iname in additional_inames:
            replace_dict[Variable(iname)] = i
        expression = substitute(pymbolic_expr, replace_dict)

        # Write timing stuff for jacobian (for alpha methods it is done at the end of stage 1)
        timer_dep = frozenset()
        if get_option("instrumentation_level") >= 4:
            timer_name = assembler_routine_name() + '_kernel' + '_quadratureloop'
            post_include('HP_DECLARE_TIMER({});'.format(timer_name), filetag='operatorfile')
            dump_accumulate_timer(timer_name)
            if(visitor.inames):
                timer_dep = frozenset({instruction(code="HP_TIMER_START({});".format(timer_name),
                                                   within_inames=frozenset(visitor.inames))})

        # Determine dependencies
        from loopy.match import Or, Writes
        from loopy.symbolic import DependencyMapper
        from dune.perftool.tools import get_pymbolic_basename
        deps = Or(tuple(Writes(get_pymbolic_basename(expr)) for expr in DependencyMapper()(expression)))

        # Issue an instruction in the quadrature loop that fills the buffer
        # with the evaluation of the contribution at all quadrature points
        assignee = Subscript(Variable(temp), tuple(Variable(i) for i in quadrature_inames()) + index)
        contrib_dep = instruction(assignee=assignee,
                                  expression=expression,
                                  forced_iname_deps=frozenset(quadrature_inames() + visitor.inames),
                                  forced_iname_deps_is_final=True,
                                  tags=frozenset({"quadvec"}).union(vectag),
                                  depends_on=frozenset({deps}).union(timer_dep)
                                  )

        if insn_dep is None:
            insn_dep = frozenset({contrib_dep})

        if get_option("instrumentation_level") >= 4:
            insn_dep = frozenset({instruction(code="HP_TIMER_STOP({});".format(timer_name),
                                              depends_on=insn_dep,
                                              within_inames=frozenset(visitor.inames))})

        inames = tuple(accum_iname((accterm.argument.restriction, restriction), mat.rows, i)
                       for i, mat in enumerate(a_matrices))

        # Collect the lfs and lfs indices for the accumulate call
        test_lfs = determine_accumulation_space(accterm.argument.expr, 0, measure)
        test_lfs.index = flatten_index(tuple(Variable(i) for i in inames),
                                       (basis_functions_per_direction(),) * dim,
                                       order="f"
                                       )

        # In the jacobian case, also determine the space for the ansatz space
        ansatz_lfs = determine_accumulation_space(accterm.term, 1, measure)
        rank = 2 if visitor.inames else 1
        if rank == 2:
            # TODO the next line should get its inames from
            # elsewhere. This is *NOT* robust (but works right now)
            ansatz_lfs.index = flatten_index(tuple(Variable(visitor.inames[i])
                                                   for i in range(world_dimension())),
                                             (basis_functions_per_direction(),) * dim,
                                             order="f"
                                             )

        # Construct the expression representing "{r,jac}.accumulate(..)"
        accum = name_accumulation_variable(test_lfs.get_restriction() + ansatz_lfs.get_restriction())

        # We can directly accumulate the solution in the last step of
        # the sumfactorization. Currently this is only implemented for
        # FastDGGridOperator.
        direct_output = None
        if get_option('fastdg'):
            ft = get_global_context_value("form_type")
            accum = accum + ".data()"
            if ft == 'residual' or ft == 'jacobian_apply':
                direct_output = accum
            else:
                assert ft == 'jacobian'
                direct_output = accum

        # Add a sum factorization kernel that implements the multiplication
        # with the test function (stage 3)
        pref_pos = i if accterm.new_indices else None
        if not get_global_context_value("dry_run", False):
            from dune.perftool.sumfact.realization import realize_sum_factorization_kernel
            result, insn_dep = realize_sum_factorization_kernel(sf,
                                                    insn_dep=insn_dep,
                                                    direct_output=direct_output,
                                                    )
        else:
            result = sf
        # Determine the expression to accumulate with. This depends on the vectorization strategy!
        result = prim.Subscript(result, tuple(prim.Variable(i) for i in inames))
        vecinames = ()
        if index:
            iname = accum_iname((accterm.argument.restriction, restriction), 4, "vec")
            vecinames = (iname,)
            transform(lp.tag_inames, [(iname, "vec")])
            from dune.perftool.tools import maybe_wrap_subscript
            result = prim.Call(prim.Variable("horizontal_add"),
                               (maybe_wrap_subscript(result, prim.Variable(iname)),),
                               )

        # In the fastdg case we accumulate directly in the last step of the SF!
        if get_option('fastdg'):
            # In the dry run we need an instruction with a
            # sumfactorization node. We use this to decide on a
            # vectorization strategy. This is just a dummy
            # instruction, in the real run the accumulation is done
            # directly in the sumfactorization.
            if get_global_context_value("dry_run", False):
                dummy = "dummy"
                globalarg(dummy, dtype=np.float64)
                instruction(assignee=prim.Variable(dummy),
                            expression=result,
                            forced_iname_deps=frozenset(inames),
                            forced_iname_deps_is_final=True,
                            depends_on=insn_dep,
                            )
        # Default: Generate accumulation instructions
        else:
            expr = Call(PDELabAccumulationFunction(accum, rank),
                        (test_lfs.get_args() +
                         ansatz_lfs.get_args() +
                         (result,)
                         )
                        )
            instruction(assignees=(),
                        expression=expr,
                        forced_iname_deps=frozenset(inames + visitor.inames + vecinames),
                        forced_iname_deps_is_final=True,
                        depends_on=insn_dep,
                        )

        # Mark the transformation that moves the quadrature loop
        # inside the trialfunction loops for application
        transform(nest_quadrature_loops, visitor.inames)

        return insn_dep

    # Extract the restrictions on argument-1:
    jac_restrictions = frozenset(tuple(ma.restriction for ma in
                                       extract_modified_arguments(accterm.term, argnumber=1, do_index=True)))
    if not jac_restrictions:
        jac_restrictions = frozenset({0})

    insn_dep = None
    for restriction in jac_restrictions:
        if accterm.new_indices:
            for i in range(world_dimension()):
                insn_dep = emit_sumfact_kernel(i, restriction, insn_dep)
        else:
            insn_dep = emit_sumfact_kernel(None, restriction, insn_dep)


    """Create a sum factorization kernel

    Sum factorization can be written as

    Y = R_{d-1} (A_{d-1} * ... * R_0 (A_0 * X)...)

    with:
    - X: Input rank d tensor of dimension n_0 x ... x n_{d-1}
    - Y: Output rank d tensor of dimension m_0 x ... x m_{d-1}
    - A_l: Values of 1D basis evaluations at quadrature points in l
      direction, matrix of dimension m_l x n_l
    - R_l: Transformation operator that permutes the underlying data
      vector of the rank d tensor in such a way that the fastest
      direction gets the slowest direction

    In the l'th step we have the following setup:
    - A_l: Matrix of dimensions m_l x n_l
    - X_l: Rank d tensor of dimensions n_l x ... x n_{d-1} x m_0 x ... x m_{l-1}
    - R_l: Transformation operator

    Looking at the indizes the following will happen:
    X --> [n_l,...,n_{d-1},m_0,...,m_{l-1}]
    A_l * X --> [m_l,n_l] * [n_l, ...] = [m_l,n_{l+1},...,n_{d-1},m_0,...,m_{l-1}]
    R_l (A_l*X) --> [n_{l+1},...,n_{d-1},m_0,...,m_{l-1}]

    So the multiplication with A_l is a reduction over one index and
    the transformation brings the next reduction index in the fastest
    position.

    Note: In the code below the transformation step is directly done
    in the reduction instruction by adapting the assignee!

    It can make sense to permute the order of directions. If you have
    a small m_l (e.g. stage 1 on faces) it is better to do direction l
    first. This can be done by:

    - Permuting the order of the A matrices.
    - Permuting the input tensor.
    - Permuting the output tensor (this assures that the directions of
      the output tensor are again ordered from 0 to d-1).

    Arguments:
    ----------
    a_matrices: An iterable of AMatrix instances
        The list of tensors to be applied to the input.
        Order of application is from 0 up.
    buf: A string identifying the flip flop buffer in use
        for intermediate results. The memory is expected to be
        pre-initialized with the input or you have to provide
        direct_input (FastDGGridOperator).
    insn_dep: An instruction ID that the first issued instruction
        should depend upon. All following ones will depend on each
        other.
    additional_inames: Instructions will be executed within those
        inames (needed for stage 3 in jacobians).
    preferred_position: Will be used in the dry run to order kernels
        when doing vectorization e.g. (dx u,dy u,dz u, u).
    outshape: Shape of the output.
    restriction: Restriction for faces values.
    direct_input: Global data structure containing input for
        sumfactorization (e.g. when using FastDGGridOperator).
    """